@@ -7,6 +7,8 @@
 #include <getopt.h>
 #include <sys/resource.h>
 #include <sys/mman.h>
+#include <signal.h>
+#include <limits.h>
 
 extern int screen_width;
 extern int screen_savebuffer_len;
@@ -36,9 +38,10 @@ fsck_progress ("Usage: %s [mode] [options] "					\
 "  -z | --adjust-size\t\tfix file sizes to real size\n"				\
 "  -q | --quiet\t\t\tno speed info\n"						\
 "  -y | --yes\t\t\tno confirmations\n"						\
+"  -f | --force\t\tforce checking even if the file system is marked clean\n"\
 "  -V\t\t\t\tprints version and exits\n"					\
 "  -a and -p\t\t\tsome light-weight auto checks for bootup\n"			\
-"  -f and -r\t\t\tignored\n"							\
+"  -r\t\t\tignored\n"							\
 "Expert options:\n"								\
 "  --no-journal-available\tdo not open nor replay journal\n"			\
 "  -S | --scan-whole-partition\tbuild tree of all blocks of the device\n\n",	\
@@ -102,6 +105,7 @@ static char * parse_options (struct fsck_data * data, int argc, char * argv [])
 	    {"adjust-size", no_argument, 0, 'z'},
 	    {"quiet", no_argument, 0, 'q'},
 	    {"yes", no_argument, 0, 'y'},
+	    {"force", no_argument, 0, 'f'},
 	    {"nolog", no_argument, 0, 'n'},
 	    
 	    /* if file exists ad reiserfs can be load of it - only
@@ -227,10 +231,13 @@ static char * parse_options (struct fsck_data * data, int argc, char * argv [])
 
 	case 'a':
 	case 'p':
+		data->options |= OPT_QUIET;
 		mode = FSCK_AUTO;
 		break;
 	
 	case 'f':
+		data->options |= OPT_FORCE;
+		break;
 	case 'r': /* ignored */
 	    break;
 	    
@@ -334,7 +341,7 @@ void warn_what_will_be_done (char * file_name, struct fsck_data * data)
 {
     FILE * warn_to;
 
-    warn_to = (data->progress ?: stderr);
+    warn_to = (data->progress ? data->progress : stderr);
 
     if (data->mode == FSCK_REBUILD)
 	reiserfs_warning (warn_to, REBUILD_WARNING, file_name);
@@ -412,7 +419,7 @@ void warn_what_will_be_done (char * file_name, struct fsck_data * data)
 
     reiserfs_warning (warn_to, "Will put log info to '%s'\n", 
 	(data->log == stdout) ? "stdout" : 
-				(data->log_file_name ? : "fsck.run"));
+	(data->log_file_name ? data->log_file_name : "fsck.run"));
     
     if (!(data->options & OPT_YES) && !user_confirmed (warn_to, "\nDo you want to "
 	"run this program?[N/Yes] (note need to type Yes if you do):", "Yes\n"))
@@ -663,6 +670,22 @@ static int where_to_start_from (reiserfs_filsys_t * fs)
     return START_FROM_THE_BEGINNING;
 }
 
+static void reiserfs_update_interval_fields(reiserfs_filsys_t *fs)
+{
+	/* Not supported on v3.5 */
+	if (get_sb_version (fs->fs_ondisk_sb) == REISERFS_FORMAT_3_5)
+		return;
+
+	set_sb_v2_lastcheck(fs->fs_ondisk_sb, time(NULL));
+	set_sb_v2_mnt_count(fs->fs_ondisk_sb, 1);
+
+	if (get_sb_v2_max_mnt_count(fs->fs_ondisk_sb) == 0)
+		set_sb_v2_max_mnt_count(fs->fs_ondisk_sb,
+					      DEFAULT_MAX_MNT_COUNT);
+	if (get_sb_v2_check_interval(fs->fs_ondisk_sb) == 0)
+		set_sb_v2_check_interval(fs->fs_ondisk_sb,
+					       DEFAULT_CHECK_INTERVAL);
+}
 
 static void mark_filesystem_consistent (reiserfs_filsys_t * fs)
 {
@@ -672,7 +695,8 @@ static void mark_filesystem_consistent (reiserfs_filsys_t * fs)
     if (!reiserfs_journal_opened (fs)) {
 	/* make sure journal is not standard */
 	if (!is_reiserfs_jr_magic_string (fs->fs_ondisk_sb))
-	    die ("Filesystem with default journal must be opened.");
+	    reiserfs_exit(EXIT_OPER, "Filesystem with default journal "
+			  "must be opened.");
 	
 	fsck_progress ("WARNING: You must use reiserfstune to specify a new "
 	    "journal before mounting it.\n");
@@ -684,6 +708,7 @@ static void mark_filesystem_consistent (reiserfs_filsys_t * fs)
 
     set_sb_umount_state (fs->fs_ondisk_sb, FS_CLEANLY_UMOUNTED);
     set_sb_fs_state (fs->fs_ondisk_sb, FS_CONSISTENT);
+    reiserfs_update_interval_fields(fs);
     
     mark_buffer_dirty (fs->fs_super_bh);
 }
@@ -784,13 +809,17 @@ static void prepare_fs_for_check(reiserfs_filsys_t * fs) {
 	/* If not CHECK mode, lock the process in the memory. */
 	if (fsck_mode (fs) != FSCK_CHECK) {
 	    if (mlockall(MCL_CURRENT)) {
-		    reiserfs_warning (stderr, "Failed to lock the process to "
-				      "fsck the mounted ro partition. %s.\n", 
-				      strerror(errno));
-		    exit(EXIT_OPER);
+		    reiserfs_exit(EXIT_OPER, "Failed to lock the process to "
+				  "fsck the mounted ro partition. %s.\n", 
+				  strerror(errno));
 	    }
 	}
 	
+	if (fsck_skip_journal (fs)) {
+		reiserfs_exit(EXIT_USER, "Jounrnal of the mounted "
+			      "filesystem must be specified.\n");
+	}
+	
 	if (!reiserfs_journal_opened (fs)) {
 	    /* just to make sure */
 	    reiserfs_panic ("Journal is not opened");
@@ -801,11 +830,6 @@ static void prepare_fs_for_check(reiserfs_filsys_t * fs) {
 	
 	fsck_progress ("Filesystem seems mounted read-only. Skipping journal "
 	    "replay.\n");
-
-	if (fsck_mode (fs) == FSCK_FIX_FIXABLE) {
-	    fsck_progress ("--fix-fixable ignored\n");
-	    fsck_mode (fs) = FSCK_CHECK;
-	}
     } else if (!fsck_skip_journal (fs)) {
 	if (reiserfs_journal_params_check(fs)) {
 	    reiserfs_close (fs);
@@ -883,7 +907,7 @@ static void clean_attributes (reiserfs_filsys_t * fs) {
     time (&t);
 
     if (get_sb_umount_state (fs->fs_ondisk_sb) != FS_CLEANLY_UMOUNTED) {
-        fsck_progress ("Filesystem does not look cleanly umounted\n"
+        fsck_progress ("Filesystem is not clean\n"
 	    "Check consistency of the partition first.\n");
         exit(EXIT_USER);
     }
@@ -922,24 +946,51 @@ static void clean_attributes (reiserfs_filsys_t * fs) {
 
 }
 
-/* Do not allow buffers to be flushed after finishing to avoid another bitmap 
- * reading on mounting. */
-static void fsck_sleep() {
-    int res;
-    
-    res = fork();
-    
-    if (res == -1) {
-	reiserfs_panic ("reiserfsck: Fork failed: %s", strerror(errno));
-    } else if (res == 0) {
-	/* Make the child process to sleep for 5 secs. */
-	sleep(5);
-    }
+static int reiserfs_check_auto_state(reiserfs_filsys_t *fs)
+{
+	time_t now = time(NULL);
+	time_t lastcheck = get_sb_v2_lastcheck(fs->fs_ondisk_sb);
+	unsigned int mnt_count = get_sb_v2_mnt_count(fs->fs_ondisk_sb);
+	unsigned int max_mnt_count = get_sb_v2_max_mnt_count(fs->fs_ondisk_sb);
+	unsigned int interval =  get_sb_v2_check_interval(fs->fs_ondisk_sb);
+
+	/* v3.5 file systems don't have the superblock fields for this */
+	if (get_sb_version (fs->fs_ondisk_sb) == REISERFS_FORMAT_3_5)
+		return 1;
+
+	if (lastcheck == 0 || mnt_count == 0 || max_mnt_count == 0 ||
+	    interval == 0) {
+		fprintf(stderr, "File system hasn't been enabled for faster "
+		        "boot-time checking. It will be enabled after a "
+			"successful run.\n");
+		return 1;
+	}
+
+	if (interval != UINT_MAX && now > lastcheck + interval) {
+		fprintf(stderr, "File system hasn't been checked in %u days. "
+		        "Checking now.\n", (now - lastcheck) / (60*60*24));
+		return 1;
+	}
+
+	if (interval != UINT_MAX && lastcheck > now) {
+		fprintf(stderr, "File system check timestamp is in the future. "
+			"Checking now.\n");
+		return 1;
+	}
+
+	if (mnt_count > max_mnt_count && max_mnt_count != USHRT_MAX) {
+		fprintf(stderr, "File system has been mounted %u times "
+		        "without being checked. Checking now.\n", mnt_count);
+		return 1;
+	}
+
+	return 0;
 }
 
 static int auto_check (reiserfs_filsys_t *fs) {
     __u16 state;
     int retval = 0;
+    int force = fsck_data(fs)->options & OPT_FORCE;
     
     print_super_block (stdout, fs, fs->fs_file_name, fs->fs_super_bh, 1);
     
@@ -961,7 +1012,12 @@ static int auto_check (reiserfs_filsys_t *fs) {
 	fprintf(stderr, "Some strange state was specified in the super block. "
 	    "Do usual check.\n");
 
+    if (get_sb_umount_state (fs->fs_ondisk_sb) == FS_CLEANLY_UMOUNTED &&
+	!force && !reiserfs_check_auto_state(fs))
+	    exit(EXIT_OK);
     prepare_fs_for_check(fs);
+    if (!force && !reiserfs_check_auto_state(fs))
+	    exit(EXIT_OK);
 
     /* Check bitmaps. */
     retval = reiserfs_open_ondisk_bitmap (fs);
@@ -1000,11 +1056,14 @@ static int auto_check (reiserfs_filsys_t *fs) {
 	/* run fixable pass. */
 	return 0;
     }
-    
+
+    reiserfs_update_interval_fields(fs);
+    mark_buffer_dirty(fs->fs_super_bh);
+    bwrite(fs->fs_super_bh);
+    fs->fs_dirt = 1;
+
     clean_after_dma_check(fs->fs_dev, &dma_info);
     
-    fsck_sleep();
-    
     reiserfs_close (fs);
     /* do not do anything else. */    
     exit (EXIT_OK);
@@ -1115,7 +1174,7 @@ static void check_fs (reiserfs_filsys_t * fs)
     reiserfs_close (fs);
     close_rollback_file ();
     
-    clear_relocated_list();    
+    //clear_relocated_list();    
     
     time (&t);
     fsck_progress ("###########\n"
@@ -1190,10 +1249,10 @@ int main (int argc, char * argv [])
 {
     char * file_name;
     struct fsck_data * data;
-    struct rlimit rlim = {0xffffffff, 0xffffffff};
+    struct rlimit rlim = {RLIM_INFINITY, RLIM_INFINITY};
     char *width;
     int retval;
-    int errno;
+    int error;
     
     width = getenv("COLUMNS");
     if ( width )
@@ -1231,8 +1290,9 @@ int main (int argc, char * argv [])
            'fsck.run'. Logs get there if log file was not specified*/
 	data->options |= OPT_QUIET;
 	data->progress = fopen ("fsck.run", "a+");
-	if (!data->progress)
-	    reiserfs_panic ("reiserfsck: Cannot not open \"fsck.run\"");
+	if (!data->progress) {
+	    reiserfs_exit(EXIT_OPER, "reiserfsck: Cannot not open \"fsck.run\"");
+	}
 
 	if (data->log == stdout)
 	    /* no log file specifed - redirect log into 'fsck.run' */
@@ -1257,21 +1317,22 @@ int main (int argc, char * argv [])
     	if (open_devices_for_rollback (file_name, data) == -1)
     	    exit(EXIT_OPER);
     } else {
-	fs = reiserfs_open (file_name, O_RDONLY, &errno, data, 
+	fs = reiserfs_open (file_name, O_RDONLY, &error, data, 
 	    data->mode != FSCK_SB);
-
+		
+	if (error) {
+		reiserfs_exit(EXIT_OPER, "Failed to open the device "
+			      "'%s': %s\n\n", file_name, strerror(error));
+	} 
+	
 	if (data->mode != FSCK_SB) {
 	    if (no_reiserfs_found (fs)) {
-		if (errno) {
-		    die ("Failed to open the device '%s': %s\n\n", 
-			 file_name, strerror(errno));
-		} else {
-		    die ("Failed to open the filesystem.\n\n"
-			 "If the partition table has not been changed, and the partition is\n"
-			 "valid  and  it really  contains  a reiserfs  partition,  then the\n"
-			 "superblock  is corrupted and you need to run this utility with\n"
-			 "--rebuild-sb.\n");
-		}
+		reiserfs_exit(EXIT_OPER, "Failed to open the filesystem.\n\n"
+			      "If the partition table has not been changed, "
+			      "and the partition is\nvalid  and  it really  "
+			      "contains  a reiserfs  partition,  then the\n"
+			      "superblock  is corrupted and you need to run "
+			      "this utility with\n--rebuild-sb.\n");
 	    }
 	    if (fsck_skip_journal (fs) && 
 		!is_reiserfs_jr_magic_string (fs->fs_ondisk_sb)) 
